import joblib
import pandas as pd
import numpy as np
from pathos.multiprocessing import ProcessPool
import multiprocessing
import inspect
from tuneta.config import *
from tuneta.optimize import Optimize
import pandas_ta as pta
from finta import TA as fta
import talib as tta
import re
from tabulate import tabulate
from tuneta.utils import col_name
from tuneta.utils import distance_correlation
from collections import OrderedDict
from scipy.spatial.distance import squareform
from joblib import delayed, Parallel
import itertools
from datetime import datetime


# Distance correlation
def dc(p0, p1):
    df = pd.concat([p0, p1], axis=1).dropna()
    res = distance_correlation(np.array(df.iloc[:, 0]).astype(float), np.array(df.iloc[:, 1]).astype(float))
    return res


class TuneTA():

    def __init__(self, n_jobs=1, verbose=False):
        self.fitted = []
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, X, y, trials=5, indicators=['tta'], ranges=[(3, 180)], early_stop=99999, min_target_correlation=0.001):
        """
        Optimize indicator parameters to maximize correlation
        :param X: Historical dataset
        :param y: Target used to measure correlation.  Can be a subset index of X
        :param trials: Number of optimization trials per indicator set
        :param indicators: List of indicators to optimize
        :param ranges: Parameter search space
        :param early_stop: Max number of optimization trials before stopping
        """
        # No missing values allowed
        if X.isna().any().any() or y.isna().any():
            raise ValueError("X and y cannot contain missing values")

        if not isinstance(X.index.get_level_values(0)[0], datetime):
            raise ValueError("Index must be of type datetime")

        if len(X) != len(y):
            raise ValueError("Length of X and y must be identical")

        self.fitted = []  # List containing each indicator completed study
        X.columns = X.columns.str.lower()  # columns must be lower case
        pool = ProcessPool(nodes=self.n_jobs)  # Set parallel cores

        # Package level optimization
        if 'tta' in indicators:
            indicators = indicators + talib_indicators
            indicators.remove('tta')
        if 'pta' in indicators:
            indicators = indicators + pandas_ta_indicators
            indicators.remove('pta')
        if 'fta' in indicators:
            indicators = indicators + finta_indicatrs
            indicators.remove('fta')
        if 'all' in indicators:
            indicators = talib_indicators + pandas_ta_indicators + finta_indicatrs
        indicators = list(OrderedDict.fromkeys(indicators))

        # Create textual representation of function in Optuna format
        # Example: 'tta.RSI(X.close, length=trial.suggest_int(\'timeperiod1\', 2, 1500))'
        # Utilizes the signature of the indicator (ie user parameters) if available
        # TTA uses help docstrings as signature is not available in C bindings
        # Parameters contained in config.py are tuned

        # Iterate user defined search space ranges
        for low, high in ranges:
            if low <= 1:
                raise ValueError("Range low must be > 1")
            if high >= len(X):
                raise ValueError(f"Range high:{high} must be > length of X:{len(X)}")

            # Iterate indicators per range
            for ind in indicators:

                # Index column to optimize if indicator returns dataframe
                idx = 0
                if ":" in ind:
                    idx = int(ind.split(":")[1])
                    ind = ind.split(":")[0]
                fn = f"{ind}("

                # If TTA indicator, use doc strings for lack of better way to
                # get indicator arguments (C binding)
                if ind[0:3] == "tta":
                    usage = eval(f"{ind}.__doc__").split(")")[0].split("(")[1]
                    params = re.sub('[^0-9a-zA-Z_\s]', '', usage).split()

                # Pandas-TA and FinTA both can be inspected for parameters
                else:
                    sig = inspect.signature(eval(ind))
                    params = sig.parameters.values()

                # Format function string
                suggest = False
                for param in params:
                    param = re.split(':|=', str(param))[0].strip()
                    if param == "open_":
                        param = "open"
                    if param == "real":
                        fn += f"X.close, "
                    elif param == "ohlc":
                        fn += f"X, "
                    elif param == "ohlcv":
                        fn += f"X, "
                    elif param in tune_series:
                        fn += f"X.{param}, "
                    elif param in tune_params:
                        suggest = True
                        fn += f"{param}=trial.suggest_int('{param}', {low}, {high}), "
                fn += ")"

                # Only optimize indicators that contain tunable parameters
                if suggest:
                    self.fitted.append(pool.apipe(Optimize(function=fn, n_trials=trials).fit, X, y, idx=idx,
                        verbose=self.verbose, early_stop=early_stop))
                else:
                    self.fitted.append(pool.apipe(Optimize(function=fn, n_trials=1).fit, X, y, idx=idx,
                        verbose=self.verbose, early_stop=early_stop))

        # Blocking wait to retrieve results
        self.fitted = [fit.get() for fit in self.fitted]

        # Fits must contain best trial data
        self.fitted = [f for f in self.fitted if len(f.study.user_attrs) > 0]

        # Remove any fits with less than minimum target correlation
        self.fitted = [f for f in self.fitted if f.study.user_attrs['best_trial'].user_attrs['correlation'] > min_target_correlation]

        # Order fits by correlation (Descending)
        self.fitted = sorted([f for f in self.fitted], key=lambda x:x.study.user_attrs['best_trial'].value, reverse=True)

    def prune(self, max_inter_correlation=.7):
        """
        Select most correlated with target, least intercorrelated
        :param top: Selects top x most correlated with target
        :param studies: From top x, keep y least intercorelated
        :return:
        """

        if not hasattr(self, 'f_corr'):
            self.features_corr()

        # Iteratively removes least fit individual of most correlated pairs of studies
        # IOW, finds most correlated pairs, removes lest correlated to target until x studies
        components = list(range(len(self.fitted)))
        indices = list(range(len(self.fitted)))
        correlations = np.array(self.f_corr)

        most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
        correlation = correlations[most_correlated[0], most_correlated[1]]
        while correlation > max_inter_correlation:
            most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
            worst = max(most_correlated)
            components.pop(worst)
            indices.remove(worst)
            correlations = correlations[:, indices][indices, :]
            indices = list(range(len(components)))
            most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
            correlation = correlations[most_correlated[0], most_correlated[1]]

        # Remove most correlated fits
        self.fitted = [self.fitted[i] for i in components]

        # Recalculate correlation of fits
        self.target_corr()
        self.features_corr()

    def transform(self, X, columns=None):
        """
        Given X, create features of fitted studies
        :param X: Dataset with features used to create fitted studies
        :return:
        """
        # Remove trailing identifier in column list if present
        if columns is not None:
            columns = [re.sub(r'_[0-9]+$', '', s) for s in columns]

        X.columns = X.columns.str.lower()  # columns must be lower case
        pool = ProcessPool(nodes=self.n_jobs)  # Number of jobs
        self.result = []

        # Iterate fitted studies and calculate TA with fitted parameter set
        for ind in self.fitted:
            # Create field if no columns or is in columns list
            if columns is None or ind.res_y.name in columns:
                self.result.append(pool.apipe(ind.transform, X))

        # Blocking wait for asynchronous results
        self.result = [res.get() for res in self.result]

        # Combine results into dataframe to return
        res = pd.concat(self.result, axis=1)
        return res

    def target_corr(self):
        fns = []  # Function names
        cor = []  # Target Correlation
        for fit in self.fitted:
            fns.append(col_name(fit.function, fit.study.user_attrs['best_trial'].params))
            cor.append(np.round(fit.study.user_attrs['best_trial'].value, 6))

        # Target correlation
        self.t_corr = pd.DataFrame(cor, index=fns, columns=['Correlation']).sort_values(by=['Correlation'], ascending=False)

    def features_corr(self):
        fns = []  # Function names
        cor = []  # Target Correlation
        features = []
        for fit in self.fitted:
            fns.append(col_name(fit.function, fit.study.user_attrs['best_trial'].params))
            cor.append(np.round(fit.study.user_attrs['best_trial'].value, 6))
            features.append(fit.study.user_attrs['best_trial'].user_attrs['res_y'])

        # Feature must be same size for correlation and of type float
        start = max([f.first_valid_index() for f in features])
        features = [(f[f.index >= start]).astype(float) for f in features]

        # Inter Correlation
        pair_order_list = itertools.combinations(features, 2)
        correlations = Parallel(n_jobs=self.n_jobs)(delayed(dc)(p[0], p[1]) for p in pair_order_list)
        correlations = squareform(correlations)
        self.f_corr = pd.DataFrame(correlations, columns=fns, index=fns)

    def report(self, target_corr=True, features_corr=True):
        if target_corr:
            if not hasattr(self, 't_corr'):
                self.target_corr()
            print("\nIndicator Correlation to Target:\n")
            print(tabulate(self.t_corr, headers=self.t_corr.columns, tablefmt="simple"))

        if features_corr:
            if not hasattr(self, 'f_corr'):
                self.features_corr()
            print("\nIndicator Correlation to Each Other:\n")
            print(tabulate(self.f_corr, headers=self.f_corr.columns, tablefmt="simple"))

    def fit_times(self):
        times = [fit.time for fit in self.fitted]
        inds = [fit.function.split('(')[0] for fit in self.fitted]
        df = pd.DataFrame({'Indicator': inds, 'Times': times}).sort_values(by='Times', ascending=False)
        print(tabulate(df, headers=df.columns, tablefmt="simple"))

    def prune_df(self, X, y, min_target_correlation=.05, max_inter_correlation=.7, report=True):
        if X.isna().any().any() or y.isna().any():
            raise ValueError("X and y cannot contain missing values")

        # Correlations to target
        tc = [distance_correlation(np.array(x[1]), np.array(y)) for x in X.iteritems()]
        names = [x[0] for x in X.iteritems()]
        target_correlation = pd.DataFrame(tc, index=names, columns=['Correlation']).sort_values(by=['Correlation'], ascending=False)

        # Columns greater than 0 correlation
        target_correlation = target_correlation[target_correlation.Correlation > min_target_correlation]

        if report:
            print("\nIndicator Correlation to Target:\n")
            print(tabulate(target_correlation, headers=target_correlation.columns, tablefmt="simple"))

        # Calculate inter correlation
        columns = target_correlation.index.values
        features = [x[1] for x in X[columns].iteritems()]
        pair_order_list = itertools.combinations(features, 2)
        correlations = Parallel(n_jobs=self.n_jobs)(delayed(dc)(p[0], p[1]) for p in pair_order_list)  # Parallelize correlation calculation
        correlations = squareform(correlations)
        components = list(range(len(correlations)))
        indices = list(range(len(correlations)))
        most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
        correlation = correlations[most_correlated[0], most_correlated[1]]
        while correlation > max_inter_correlation:
            most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
            worst = max(most_correlated)
            components.pop(worst)
            indices.remove(worst)
            correlations = correlations[:, indices][indices, :]
            indices = list(range(len(components)))
            most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
            correlation = correlations[most_correlated[0], most_correlated[1]]

        # Get columns of features to keep
        columns = columns[components]

        # Report intercorrelation
        if report:
            correlations = pd.DataFrame(correlations, columns=columns, index=columns)
            print("\nIntercorrelation after prune:\n")
            print(tabulate(correlations, headers=correlations.columns, tablefmt="simple"))

        return columns





# from tuneta.optimize import _weighted_spearman
import pandas as pd
from tabulate import tabulate
import numpy as np
import re
from scipy.spatial.distance import squareform, pdist
import dcor


def col_name(function, study_best_params):
    """
    Create consistent column names given string function and params
    :param function:  Function represented as string
    :param study_best_params:  Params for function
    :return:
    """

    # Optuna string of indicator
    function_name = function.split("(")[0].replace(".", "_")

    # Optuna string of parameters
    params = re.sub('[^0-9a-zA-Z_:,]', '', str(study_best_params)).replace(",", "_").replace(":", "_")

    # Concatenate name and params to define
    col = f"{function_name}_{params}"

    # Remove any trailing underscores
    col = re.sub(r'_$', '', col)
    return col


def distance_correlation(a, b):
    return dcor.distance_correlation(a, b)


# import seaborn as sns
# import matplotlib.pyplot as plt

def gen_plot(indicators, title):
    data = pd.DataFrame()
    for fitted in indicators.fitted:
        fitted.fitness = []
        fitted.length = []
        for trial in fitted.study.trials:
            print(trial)
            fitted.fitness.append(trial.value)
            fitted.length.append(trial.params['length'])
        fitted.fitness = pd.Series(fitted.fitness, name="Correlation")
        fitted.length = pd.Series(fitted.length, name="Length")
        fitted.data = pd.DataFrame([fitted.fitness, fitted.length]).T
        fitted.fn = fitted.function.split('(')[0]
        fitted.data['Indicator'] = fitted.fn
        data = pd.concat([data, fitted.data])
        fitted.x = fitted.study.best_params['length']
        fitted.y = fitted.study.best_value

    plt.figure(figsize=(10, 6))
    sns.scatterplot(x="Length", y="Correlation", data=data, hue="Indicator")
    plt.title(title)
    for fit in indicators.fitted:
        plt.vlines(x=fit.x, ymin=0, ymax=fit.y, linestyles='dotted')




# Series data
tune_series = ['open', 'high', 'low', 'close', 'volume']

# Parameters to tune
tune_params = ['width', 'fast', 'slow', 'signal', 'length', 'max_lookback', 'min_lookback', 'initial',
               'lower_length', 'upper_length', 'lookback', 'medium', 'slow_w', 'roc1', 'roc2', 'roc3', 'roc4',
               'sma1', 'sma2', 'sma3', 'sma4', 'scalar', 'long', 'atr_length', 'high_length',
               'low_length', 'kc_length', 'bb_length', 'mom_length', 'swma_length', 'rvi_length', 'period',
               'er', 'ema_fast', 'ema_slow', 'period_fast', 'period_slow', 'rsi_period', 'wma_period', 'atr_period',
               'upper_period', 'lower_period', 'stoch_period', 'slow_period', 'fast_period', 'r1', 'r2', 'r3', 'r4',
               'short', 'signal', 'short_period', 'long_period', 'channel_lenght', 'channel_length',
               'average_lenght', 'average_length', 'tenkan_period', 'kijun_period', 'senkou_period', 'chikou_period',
               'k_period', 'timeperiod', 'timeperiod1', 'timeperiod2', 'timeperiod3', 'slowlimit', 'fastlimit',
               'minperiod', 'maxperiod', 'acceleration', 'accelerationlong', 'accelerationshort', 'signalperiod',
               'slowperiod', 'signalperiod', 'fastperiod', 'fastk_period', 'slowk_period', 'slowd_period',
               'fastd_period', 'k', 'd', 'smooth_k', 'p']

talib_indicators = ['tta.BBANDS', 'tta.DEMA', 'tta.EMA', 'tta.HT_TRENDLINE', 'tta.KAMA', 'tta.MA',
                        'tta.MIDPOINT', 'tta.MIDPRICE', 'tta.SAR', 'tta.SAREXT', 'tta.SMA', 'tta.T3', 'tta.TEMA',
                        'tta.TRIMA',
                        'tta.WMA', 'tta.ADX', 'tta.ADXR', 'tta.APO', 'tta.AROON:1', 'tta.AROONOSC', 'tta.BOP',
                        'tta.CCI', 'tta.CMO',
                        'tta.DX', 'tta.MACD', 'tta.MACDEXT', 'tta.MACDFIX', 'tta.MFI', 'tta.MINUS_DI', 'tta.MINUS_DM',
                        'tta.MOM',
                        'tta.PLUS_DI', 'tta.PLUS_DM', 'tta.PPO', 'tta.ROC', 'tta.ROCP', 'tta.ROCR', 'tta.ROCR100',
                        'tta.RSI', 'tta.STOCH',
                        'tta.STOCHF', 'tta.STOCHRSI', 'tta.TRIX', 'tta.ULTOSC', 'tta.WILLR', 'tta.AD', 'tta.ADOSC',
                        'tta.OBV',
                        'tta.HT_DCPERIOD', 'tta.HT_DCPHASE', 'tta.HT_PHASOR', 'tta.HT_SINE', 'tta.HT_TRENDMODE',
                        'tta.AVGPRICE', 'tta.MEDPRICE',
                        'tta.TYPPRICE', 'tta.WCLPRICE', 'tta.ATR', 'tta.NATR', 'tta.TRANGE', 'tta.CDL2CROWS',
                        'tta.CDL3BLACKCROWS',
                        'tta.CDL3INSIDE', 'tta.CDL3LINESTRIKE', 'tta.CDL3OUTSIDE', 'tta.CDL3STARSINSOUTH',
                        'tta.CDL3WHITESOLDIERS',
                        'tta.CDLABANDONEDBABY', 'tta.CDLADVANCEBLOCK', 'tta.CDLBELTHOLD', 'tta.CDLBREAKAWAY',
                        'tta.CDLCLOSINGMARUBOZU',
                        'tta.CDLCONCEALBABYSWALL', 'tta.CDLCOUNTERATTACK', 'tta.CDLDARKCLOUDCOVER', 'tta.CDLDOJI',
                        'tta.CDLDOJISTAR',
                        'tta.CDLDRAGONFLYDOJI', 'tta.CDLENGULFING', 'tta.CDLEVENINGDOJISTAR', 'tta.CDLEVENINGSTAR',
                        'tta.CDLGAPSIDESIDEWHITE',
                        'tta.CDLGRAVESTONEDOJI', 'tta.CDLHAMMER', 'tta.CDLHANGINGMAN', 'tta.CDLHARAMI',
                        'tta.CDLHARAMICROSS',
                        'tta.CDLHIGHWAVE', 'tta.CDLHIKKAKE', 'tta.CDLHIKKAKEMOD', 'tta.CDLHOMINGPIGEON',
                        'tta.CDLIDENTICAL3CROWS',
                        'tta.CDLINNECK', 'tta.CDLINVERTEDHAMMER', 'tta.CDLKICKING', 'tta.CDLKICKINGBYLENGTH',
                        'tta.CDLLADDERBOTTOM',
                        'tta.CDLLONGLEGGEDDOJI', 'tta.CDLLONGLINE', 'tta.CDLMARUBOZU', 'tta.CDLMATCHINGLOW',
                        'tta.CDLMATHOLD',
                        'tta.CDLMORNINGDOJISTAR', 'tta.CDLMORNINGSTAR', 'tta.CDLONNECK', 'tta.CDLPIERCING',
                        'tta.CDLRICKSHAWMAN',
                        'tta.CDLRISEFALL3METHODS', 'tta.CDLSEPARATINGLINES', 'tta.CDLSHOOTINGSTAR', 'tta.CDLSHORTLINE',
                        'tta.CDLSPINNINGTOP',
                        'tta.CDLSTALLEDPATTERN', 'tta.CDLSTICKSANDWICH', 'tta.CDLTAKURI', 'tta.CDLTASUKIGAP',
                        'tta.CDLTHRUSTING', 'tta.CDLTRISTAR',
                        'tta.CDLUNIQUE3RIVER', 'tta.CDLUPSIDEGAP2CROWS', 'tta.CDLXSIDEGAP3METHODS', 'tta.LINEARREG',
                        'tta.LINEARREG_ANGLE', 'tta.LINEARREG_INTERCEPT', 'tta.LINEARREG_SLOPE', 'tta.STDDEV',
                        'tta.TSF', 'tta.VAR']

pandas_ta_indicators = [
    'pta.alma',
    'pta.dema',
    'pta.ema',
    'pta.fwma',
    'pta.hilo',
    'pta.hl2',
    'pta.hlc3',
    # 'pta.hma',  # does not accept low ranges
    'pta.hwma',
    'pta.ichimoku',
    'pta.jma',
    'pta.kama',
    'pta.linreg',
    'pta.mcgd',
    'pta.midpoint',
    'pta.midprice',
    'pta.ohlc4',
    'pta.pwma',
    'pta.rma',
    'pta.sinwma',
    'pta.sma',
    'pta.ssf',
    'pta.supertrend',
    'pta.swma',
    'pta.t3',
    'pta.tema',
    'pta.trima',
    'pta.vidya',
    'pta.vwap',
    'pta.wcp',
    'pta.wma',
    'pta.zlma',
    'pta.decreasing',
    'pta.increasing',
    'pta.stdev',
    'pta.mad',
    'pta.quantile',
    'pta.skew',
    'pta.variance',
    'pta.zscore',
    'pta.median',
    'pta.kurtosis',
    'pta.entropy',
    'pta.eri',
    'pta.bias',
    'pta.cg',
    'pta.willr',
    'pta.psl',
    'pta.rvgi',
    'pta.kdj',
    'pta.smi',
    'pta.fisher',
    'pta.macd',
    'pta.mom',
    'pta.stochrsi',
    'pta.apo',
    'pta.squeeze_pro',
    'pta.inertia',
    'pta.cmo',
    'pta.slope',
    'pta.roc',
    'pta.rsx',
    'pta.dm',
    'pta.cci',
    'pta.cti',
    'pta.er',
    'pta.uo',
    'pta.bop',
    'pta.cfo',
    'pta.stc',
    'pta.tsi',
    'pta.ao',
    'pta.stoch',
    'pta.qqe',
    'pta.coppock',
    'pta.rsi',
    'pta.trix',
    'pta.ppo',
    'pta.pgo',
    'pta.squeeze',
    'pta.brar',
    'pta.td_seq',
    'pta.kst',
    'pta.pvo',
    'pta.amat',
    'pta.adx',
    'pta.vortex',
    'pta.qstick',
    'pta.psar',
    'pta.decay',
    'pta.cksp',
    'pta.chop',
    'pta.ttm_trend',
    'pta.vhf',
    'pta.aroon',
    'pta.dpo',
    'pta.donchian',
    'pta.accbands',
    'pta.ui',
    'pta.hwc',
    'pta.natr',
    'pta.atr',
    'pta.true_range',
    'pta.massi',
    'pta.thermo',
    'pta.aberration',
    'pta.kc',
    'pta.bbands',
    'pta.pdist',
    'pta.rvi',
    'pta.pvt',
    'pta.efi',
    'pta.adosc',
    'pta.aobv',
    'pta.kvo',
    'pta.pvi',
    'pta.nvi',
    'pta.vp',
    'pta.mfi',
    'pta.obv',
    'pta.pvol',
    'pta.eom',
    'pta.pvr',
    'pta.ad',
    'pta.cmf'
]

finta_indicatrs = [
    'fta.SMA',
    'fta.SMM',
    'fta.SSMA',
    'fta.EMA',
    'fta.DEMA',
    'fta.TEMA',
    'fta.TRIMA',
    'fta.TRIX',
    'fta.VAMA',
    'fta.ER',
    'fta.KAMA',
    'fta.ZLEMA',
    'fta.WMA',
    'fta.HMA',
    'fta.EVWMA',
    'fta.VWAP',
    'fta.SMMA',
    # 'fta.FRAMA',  # Requires even parameters (not implemented)
    'fta.MACD',
    'fta.PPO',
    'fta.VW_MACD',
    'fta.EV_MACD',
    'fta.MOM',
    'fta.ROC',
    'fta.RSI',
    'fta.IFT_RSI',
    'fta.TR',
    'fta.ATR',
    'fta.SAR',
    'fta.BBANDS',
    'fta.BBWIDTH',
    'fta.MOBO',
    'fta.PERCENT_B',
    'fta.KC',
    'fta.DO',
    'fta.DMI',
    'fta.ADX',
    'fta.PIVOT',
    'fta.PIVOT_FIB',
    'fta.STOCH',
    'fta.STOCHD',
    'fta.STOCHRSI',
    'fta.WILLIAMS',
    'fta.UO',
    'fta.AO',
    'fta.MI',
    'fta.VORTEX',
    'fta.KST',
    'fta.TSI',
    'fta.TP',
    'fta.ADL',
    'fta.CHAIKIN',
    'fta.MFI',
    'fta.OBV',
    'fta.WOBV',
    'fta.VZO',
    'fta.PZO',
    'fta.EFI',
    'fta.CFI',
    'fta.EBBP',
    'fta.EMV',
    'fta.CCI',
    'fta.COPP',
    'fta.BASP',
    'fta.BASPN',
    'fta.CMO',
    'fta.CHANDELIER',
    'fta.QSTICK',
    # 'fta.TMF',  # Not implemented
    'fta.WTO',
    'fta.FISH',
    'fta.ICHIMOKU',
    'fta.APZ',
    'fta.SQZMI',
    'fta.VPT',
    'fta.FVE',
    'fta.VFI',
    'fta.MSD',
    'fta.STC',
    # 'fta.WAVEPM'  # No attribute error
]




import optuna
import pandas as pd
import numpy as np
import pandas_ta as pta
from finta import TA as fta
import talib as tta
from sklearn.cluster import KMeans
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics.pairwise import euclidean_distances
from optuna.trial import TrialState
from timeit import default_timer as timer
from tuneta.utils import col_name
from tuneta.utils import distance_correlation
from yellowbrick.cluster import KElbowVisualizer
import json
import warnings
warnings.filterwarnings("ignore")


# Apply trial on multi index
def trial_results(X, function, trial, sym=None):
    if sym:
        X = X.droplevel(1)
    try:
        res = eval(function)  # Eval contains reference to best trial (in argument) to re-use original parameters
    except Exception as e:
        raise Exception(e)
    if isinstance(res, tuple):
        try:
            res = pd.DataFrame(res).T
        except Exception as e:
            print("Error:")
            print(f"Function: {function}")
            print(f"X Length:  {len(X)}")
            for k, v in enumerate(res):
                u, c = np.unique(v.index, return_counts=True)
                dup = u[c > 1]
                print(f"Series {k} duplicates: {len(dup)}")
                print(v)
            raise Exception(e)

    res = pd.DataFrame(res, index=X.index)  # Ensure result aligns with X
    if sym:
        res['sym'] = sym
        res.set_index('sym', append=True, inplace=True)
    return res


def _trial(self, trial, X):
    """
    Calculate indicator using best fitted trial over X
    :param self:  Optuna study
    :param trial:  Optuna trial
    :param X:  dataset
    :return:
    """
    if X.index.nlevels == 2:  # support 2 level inddex (data/symbol)
        res = [trial_results(X, self.function, trial, sym=sym) for sym, X in X.groupby(level=1)]
        res = pd.concat(res, axis=0).sort_index()
    else:
        res = trial_results(X, self.function, trial)

    # Create consistent column names with function string and params
    name = col_name(self.function, trial.params)

    # Append integer identifier to DF with multiple columns
    if len(res.columns) > 1:
        res.columns = [f"{name}_{i}" for i in range(len(res.columns))]
    else:
        res.columns = [f"{name}"]
    return res


def _early_stopping_opt(study, trial):
    """
    Callback function to stop optuna trials early for single objective
    :param study:  Optuna study
    :param trial:  Optuna trial
    :return:
    """

    # Don't early stop on trial 0, also avoids ValueError when accessing study too soon
    if trial.number == 0:
        return

    if len([t.state for t in study.trials if t.state == TrialState.COMPLETE]) == 0:
        return

    # Record first as best score
    if study.best_score is None:
        study.best_score = study.best_value

    # Record better value and reset count
    if study.best_value > study.best_score:
        study.best_score = study.best_value
        study.early_stop_count = 0

    # If count greater than user defined stop, raise error to stop
    else:
        if study.early_stop_count > study.early_stop:
            study.early_stop_count = 0
            raise optuna.exceptions.OptunaError
        else:
            study.early_stop_count = study.early_stop_count+1
    return


# Apply best trial parameters on multi-index dataframe
def eval_res(X, function, idx, trial, sym=None):
    if sym:
        X = X.droplevel(1)
    try:
        res = eval(function)
    except Exception as e:
        print(f"Error:  Function: {function}  Parameters: {trial.params}")
        raise Exception(e)
    if isinstance(res, tuple):
        res = res[idx]
    res = pd.DataFrame(res, index=X.index)
    if len(res.columns) > 1:
        res = pd.DataFrame(res.iloc[:, idx])
    if sym:
        res['sym'] = sym
        res.set_index('sym', append=True, inplace=True)
    return res


def _objective(self, trial, X, y):
    """
    Objective function used in Optuna trials
    :param self:  Optuna study
    :param trial:  Optuna trial
    :param X: Entire dataset
    :param y: Target
    :return:
    """

    # Execute trial function
    # try:
    if X.index.nlevels == 2:
        res = [eval_res(X, self.function, self.idx, trial, sym=sym) for sym, X in X.groupby(level=1)]
        res = pd.concat(res, axis=0).sort_index()
    else:
        res = eval_res(X, self.function, self.idx, trial)

    # y may be a subset of X, so reduce result to y and convert to series
    res_y = res.reindex(y.index).iloc[:, 0].replace([np.inf, -np.inf], np.nan)

    # Obtain distance correlation
    if sum(np.isnan(res_y)) >= len(res_y)*.9:  # If mostly nans in result, return nan correlation
        correlation = np.nan
    else:
        # Ensure results and target are aligned with target by index
        res_tgt = pd.concat([res_y, y], axis=1)
        res_tgt.columns = ['results', 'target']

        # Measure Correlation
        fvi = res_tgt['results'].first_valid_index()
        res_tgt = res_tgt[res_tgt.index >= fvi]

        if res_tgt.results.isnull().sum() > int(len(res_tgt.results) * .90):
            print(f"Error:  Function: {self.function}  Parameters: {trial.params}")
            raise ValueError(f"Excessive NANs - Function: {self.function}  Parameters: {trial.params}")
        else:
            res_tgt.dropna(inplace=True)
        correlation = distance_correlation(np.array(res_tgt.target), np.array(res_tgt.results))

    # Save results
    trial.set_user_attr("correlation", correlation)
    trial.set_user_attr("res_y", res_y)

    return correlation


class Optimize():
    def __init__(self, function, n_trials=100):
        self.function = function
        self.n_trials = n_trials

    def fit(self, X, y, idx=0, verbose=False, early_stop=None):
        """
        Optimize a technical indicator
        :param X: Historical dataset
        :param y: Target
        :param idx: Column to maximize if TA returnes multiple
        :param verbose: Verbosity level
        :param early_stop: Number of trials to perform without improvement before stopping
        :return:
        """

        start_time = timer()
        self.idx = idx

        # Display Optuna trial messages
        if not verbose:
            optuna.logging.set_verbosity(optuna.logging.ERROR)

        # Create optuna study maximizing correlation
        sampler = optuna.samplers.TPESampler(seed=123)
        self.study = optuna.create_study(direction='maximize', study_name=self.function, sampler=sampler)

        # Set required early stopping variables
        self.study.early_stop = early_stop
        self.study.early_stop_count = 0
        self.study.best_score = None

        # Start optimization trial
        try:
            self.study.optimize(lambda trial: _objective(self, trial, X, y),
                n_trials=self.n_trials, callbacks=[_early_stopping_opt], n_jobs=1)

        # Early stopping (not officially supported by Optuna)
        except optuna.exceptions.OptunaError:
            pass

        if len([t for t in self.study.trials if t.state == TrialState.COMPLETE]) == 0:  # Min 1 complete trial
            return self
        elif self.n_trials == 1:  # Indicators with no parameters
            best_trial = 0
        else:
            # Unique trials (converts params to json for duplicate comparison)
            trials = pd.DataFrame([[t.number, t.user_attrs['correlation'], t.params, json.dumps(t.params)] for t in self.study.trials if t.state == TrialState.COMPLETE])
            trials.columns = ['trial', 'correlation', 'params', 'json']
            trials.set_index('trial', drop=True, inplace=True)
            trials = trials[~trials.json.duplicated(keep='first')]
            trials = trials.drop(columns=['json'])

            # Scaler for cluster scoring
            mms = MinMaxScaler()

            # Trial correlations
            correlations = np.array(list(trials.correlation)).reshape(-1, 1)

            # Clusters of trial correlations
            if len(correlations) <= 7:
                num_clusters = 1
            else:
                max_clusters = int(min([20, len(correlations)/2]))
                ke = KElbowVisualizer(KMeans(random_state=123), k=(1, max_clusters))
                ke.fit(correlations)
                num_clusters = ke.elbow_value_
                if num_clusters is None:
                    num_clusters = int(len(correlations) * .2)
            kmeans = KMeans(n_clusters=num_clusters, random_state=123).fit(correlations.reshape(-1, 1))

            # Mean correlation per cluster, membership and score
            cluster_mean_correlation = [np.mean(trials[(kmeans.labels_ == c)].correlation) for c in range(num_clusters)]
            cluster_members = [(kmeans.labels_ == c).sum() for c in range(num_clusters)]
            clusters = pd.DataFrame([cluster_mean_correlation, cluster_members]).T
            clusters.columns = ['mean_correlation', 'members']

            # Choose best cluster
            df = pd.DataFrame(mms.fit_transform(clusters), index=clusters.index, columns=clusters.columns)
            clusters['score'] = df['mean_correlation'] + df['members']
            clusters = clusters.sort_values(by='score', ascending=False)
            cluster = clusters.score.idxmax()

            # Trials of best cluster
            trials = trials[kmeans.labels_ == cluster]

            # Trial parameters of best cluster
            num_params = len(self.study.best_params)
            params = []
            for i in range(0, num_params):
                params.append([list(p.values())[i] for p in trials.params])
            params = np.nan_to_num(np.array(params).T)

            if len(params) <= 7:
                num_clusters = 1
            else:
                # Clusters of trial parameters for best correlation cluster
                max_clusters = int(min([20, len(params)/2]))
                ke = KElbowVisualizer(KMeans(random_state=123), k=(1, max_clusters))
                ke.fit(params)
                num_clusters = ke.elbow_value_
                if num_clusters is None:
                    num_clusters = int(len(params) * .2)
            kmeans = KMeans(n_clusters=num_clusters, random_state=123).fit(params)

            # Mean correlation per cluster, membership and score
            cluster_mean_correlation = [np.mean(trials[(kmeans.labels_ == c)].correlation) for c in range(num_clusters)]
            cluster_members = [(kmeans.labels_ == c).sum() for c in range(num_clusters)]
            clusters = pd.DataFrame([cluster_mean_correlation, cluster_members]).T
            clusters.columns = ['mean_correlation', 'members']

            # Choose best cluster
            df = pd.DataFrame(mms.fit_transform(clusters), index=clusters.index, columns=clusters.columns)
            clusters['score'] = df['mean_correlation'] + df['members']
            clusters = clusters.sort_values(by='score', ascending=False)
            cluster = clusters.score.idxmax()

            # Choose center of cluster
            center = kmeans.cluster_centers_[cluster]
            center_matrix = np.vstack((center, params))
            distances = pd.DataFrame(euclidean_distances(center_matrix)[1:, :], index=trials.index)[0]
            best_trial = distances.sort_values().index[0]

            # look = trials[kmeans.labels_ == cluster]

        self.study.set_user_attr("best_trial_number", best_trial)
        self.study.set_user_attr("best_trial", self.study.trials[best_trial])
        self.study.set_user_attr("name", col_name(self.function, self.study.trials[best_trial].params))

        end_time = timer()
        self.time = round(end_time - start_time, 2)

        return self

    def transform(self, X):
        """
        Calculate TA indicator using fittted best trial
        :param X: Dataset
        :return:
        """

        # Calculate and store in features, replacing any potential non-finites (not sure needed)
        features = _trial(self, self.study.user_attrs['best_trial'], X)
        return features
